/*
   setjmp.s

   Copyright (C) 1999 Marek Michalkiewicz <marekm@linux.org.pl>

   Permission to use, copy, modify, and distribute this software and
   its documentation for any purpose and without fee is hereby granted,
   without any conditions or restrictions.  This software is provided
   "as is" without express or implied warranty.
 */

/*
   jmp_buf:
	offset	size	description
	 0	16	call-saved registers (r2-r17)
	16	 2	frame pointer (r29:r28)
	18	 2	stack pointer (SPH:SPL)
	20	 1	status register (SREG)
	21	 3	return address (PC) (2 bytes used for <=128K flash)
	24 = total size

int setjmp(jmp_buf __jmpb);
void longjmp(jmp_buf __jmpb, int __val) __attribute__((noreturn));
 */

#include "macros.inc"

/* the same library is used for 2313 and 8515 for now -
   I hope writing 0 to non-existent SPH doesn't hurt... */
#ifndef SPH
#define SPH (SPL+1)
#endif

#if BIG_PTR
  #define jmpb_hhi	rP0
  #define jmpb_hi	rP1
  #define jmpb_lo	rP2
  #define val_hi	rP3
  #define val_lo	rP4
#else
  #define jmpb_hi	rP0
  #define jmpb_lo	rP1
  #define val_hi	rP2
  #define val_lo	rP3
#endif

#define ret_lo rP1  /* r24 */
#define ret_hi rP0  /* r25 */

	.section .text

	.global _U(setjmp)
	.global _U(longjmp)

_U(setjmp):
	LOAD_Z(jmpb_lo, jmpb_hi)
	OUT_RAMPZ(jmpb_hhi)
	in	XL, SPL
	in	XH, SPH
	OUT_RAMPX(__zero_reg__)  /* stack is always in first 64K */
	/* save program counter (return address) */
	/* return address on stack (pushed by "call") is big endian! */
	/* SP is post-decremented by "call" */
	adiw	XL, 1
#ifdef EIND  /* devices with >128K bytes of flash - none yet */
	ld	__tmp_reg__, X+
	std	Z+23, __tmp_reg__
#endif
	ld	__tmp_reg__, X+
	std	Z+22, __tmp_reg__
	ld	__tmp_reg__, X
	std	Z+21, __tmp_reg__
	/* save stack pointer (SP value before calling this function) */
	std	Z+18, XL
	std	Z+19, XH
	/* save status register (I flag) */
	in	__tmp_reg__, SREG
	std	Z+20, __tmp_reg__
	/* save call-saved registers */
	st	Z, r2
	std	Z+1, r3
	std	Z+2, r4
	std	Z+3, r5
	std	Z+4, r6
	std	Z+5, r7
	std	Z+6, r8
	std	Z+7, r9
	std	Z+8, r10
	std	Z+9, r11
	std	Z+10, r12
	std	Z+11, r13
	std	Z+12, r14
	std	Z+13, r15
	std	Z+14, r16
	std	Z+15, r17
	/* save frame pointer */
	std	Z+16, YL
	std	Z+17, YH
	/* return zero */
	clr	ret_lo
	clr	ret_hi
	ret

_U(longjmp):
	LOAD_X(jmpb_lo, jmpb_hi)
	OUT_RAMPX(jmpb_hhi)
	/* return value */
	mov	ret_lo, val_lo
	mov	ret_hi, val_hi
	/* if zero, change to 1 */
	cpi	ret_lo, 1
	cpc	ret_hi, __zero_reg__
	adc	ret_lo, __zero_reg__
	/* restore call-saved registers */
	ld	r2, X+
	ld	r3, X+
	ld	r4, X+
	ld	r5, X+
	ld	r6, X+
	ld	r7, X+
	ld	r8, X+
	ld	r9, X+
	ld	r10, X+
	ld	r11, X+
	ld	r12, X+
	ld	r13, X+
	ld	r14, X+
	ld	r15, X+
	ld	r16, X+
	ld	r17, X+
	/* restore frame pointer */
	ld	YL, X+
	ld	YH, X+
	/* restore stack pointer (SP value before the setjmp() call) */
	ld	ZL, X+
	ld	ZH, X+
	ld	__tmp_reg__, X+
	/* interrupts disabled for shortest possible time (3 cycles) */
	cli
	out	SPL, ZL
	out	SPH, ZH
	/* restore status register (including the interrupt enable flag) */
	out	SREG, __tmp_reg__
	/* restore return address */
	ld	ZL, X+
	ld	ZH, X+

#if NO_IJMP
	push	ZL
	push	ZH
#ifdef EIND
	ld	__tmp_reg__, X
	push	__tmp_reg__
#endif
	ret

#else  /* use IJMP */

#ifdef EIND
	ld	__tmp_reg__, X
	out	EIND, __tmp_reg__
	.word	0x9419  /* eijmp */
#else
	ijmp
#endif
#endif /* use IJMP */

